--****************************************************************************************************
-- Client Lua Script for TargetSpy
-- Copyright (c) NCsoft. All rights reserved
-- Created by SpielePhilosoph - http://www.curse.com/users/SpielePhilosoph/projects
-- Last modified: 20140715
-- Version: 0.8
--****************************************************************************************************

--****************************************************************************************************
-- List of used libs, that need for this addon.
--****************************************************************************************************
require "Apollo"
require "GameLib"
require "Window"

--****************************************************************************************************
-- TargetSpy Module Definition
--****************************************************************************************************
local TargetSpy = {} 

--****************************************************************************************************
-- TargetSpy Variables
--****************************************************************************************************
local ConfigMinVersion = 1 -- Minumum version for config data to load it.

--****************************************************************************************************
-- Initialization
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- TargetSpy new
------------------------------------------------------------------------------------------------------
function TargetSpy:new(o)
 o = o or {}
 setmetatable(o, self)
 self.__index = self

 o.Config = {} -- Property to saved and load config data.
 o.Config.IsVisible = false -- Per default the main window is not shown.
 o.Config.PositionMain = -- Default position for TargetSpyHoloForm.
 {
  Bottom = 460,
  Left = 100,
  Right = 400,
  Top = 100,
 }
 o.Config.Version = 1 -- Version of config.
 return o
end

------------------------------------------------------------------------------------------------------
-- TargetSpy Init
------------------------------------------------------------------------------------------------------
function TargetSpy:Init()
 local bHasConfigureFunction = false
 local strConfigureButtonText = ""
 local tDependencies = {}
 Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
--****************************************************************************************************
-- Eventfunctions for TargetSpy
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- TargetSpy OnDocLoaded
------------------------------------------------------------------------------------------------------
function TargetSpy:OnDocLoaded()
 if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
  self.MainForm = Apollo.LoadForm(self.xmlDoc, "TargetSpyHoloForm", nil, self)
  if self.MainForm == nil then
   Apollo.AddAddonErrorText(self, "Could not load the holo window for some reason.")
  else
   self.MainForm:Show(false, true)
   if self.Config ~= nil then
    -- Restore saved window positions.
    if self.Config.PositionMain ~= nil then
     self.MainForm:SetAnchorOffsets(
      self.Config.PositionMain.Left,
      self.Config.PositionMain.Top,
      self.Config.PositionMain.Right,
      self.Config.PositionMain.Bottom)
    end
   end

   -- The xmlDoc is no longer needed.
   self.xmlDoc = nil
   -- View can be open via the command /targetspy or menu entry.
   Apollo.RegisterSlashCommand("targetspy", "OnTargetSpyCall", self)
   Apollo.RegisterEventHandler("TargetSpyMenuClicked", "OnTargetSpyCall", self)
   -- Register to add entries to window management.
   Apollo.RegisterEventHandler("WindowManagementReady", "OnWindowManagementReady", self)

   -- Open the main form if it was visible before.
   if self.Config ~= nil and
      self.Config.IsVisible == true then
    self:MainShow()
   end
  end
 end
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnInterfaceMenuListHasLoaded
-- Adds a menu entry for this addon.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnInterfaceMenuListHasLoaded()
 Event_FireGenericEvent(
  "InterfaceMenuList_NewAddOn",
  "Target Spy",
  {
   "TargetSpyMenuClicked",
   "",
   "IconSprites:Icon_Windows32_UI_CRB_InterfaceMenu_Abilities"
  })
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnLoad
-- Load the form xml file.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnLoad()
 self.xmlDoc = XmlDoc.CreateFromFile("TargetSpy.xml")
 self.xmlDoc:RegisterCallback("OnDocLoaded", self)
 -- Register for the InterfaceMenuListHasLoaded event to add a own menu item.
 Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnRestore
-- Load the saved settings for the addon.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnRestore(type, savedData)
 if type == GameLib.CodeEnumAddonSaveLevel.Character then
  if savedData ~= nil and
     savedData.Version ~= nil and
     savedData.Version >= ConfigMinVersion then
   -- Saved configs only used if they compatible with the minimum version.
   self.Config = savedData
  end
 end
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnSave
-- Save the settings for the addon per Character.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnSave(eType)
 local result = nil
 if eType == GameLib.CodeEnumAddonSaveLevel.Character then
  result = self.Config
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnTargetSpyCall
-- Function on SlashCommand "/targetspy"
------------------------------------------------------------------------------------------------------
function TargetSpy:OnTargetSpyCall()
 if self.MainForm:IsShown() == true then
  self:MainHide()
 else
  self:MainShow()
 end
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnTargetUnitChanged
-- Update all informations with the target values.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnTargetUnitChanged(target)
 -- Define list to use different functions
 local attributes = nil
 -- Remove all old information.
 self.MainForm:FindChild("AttributeGrid"):DeleteAll()
 if target == nil then
  target = GameLib.GetPlayerUnit()
  self.MainForm:FindChild("Title"):SetText("No target selected. (Self:" .. target:GetName() .. ")")
  attributes = self:GetAttributesSelf(target)
 else
  self.MainForm:FindChild("Title"):SetText(target:GetName())
  attributes = self:GetAttributes(target)
 end
 -- Show the attributes.
 local attributeGridHolo = self.MainForm:FindChild("AttributeGrid")
 local heightItemList = 0 
 for index, item in pairs(attributes) do
  for innerIndex, attribute in pairs(attributes[index]) do
   local row = attributeGridHolo:AddRow("")
   attributeGridHolo:SetCellImage(row, 1, attribute.icon)
   attributeGridHolo:SetCellText(row, 2, attribute.name)
   attributeGridHolo:SetCellText(row, 3, attribute.value)
  end
 end
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnWindowManagementReady
-- Add form to addon window management.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnWindowManagementReady()
 Event_FireGenericEvent("WindowManagementAdd", {wnd = self.MainForm, strName = "Target Spy"})
end

--****************************************************************************************************
-- Eventfunctions for TargetSpyHoloForm
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- TargetSpy OnClose
-- Remove handler for update target info and hide window.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnClose(wndHandler, wndControl, eMouseButtons)
 self:MainHide()
end

------------------------------------------------------------------------------------------------------
-- TargetSpy OnWindowMoveMain
-- Remove handler for update target info and hide window.
------------------------------------------------------------------------------------------------------
function TargetSpy:OnWindowMoveMain( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
 self.Config.PositionMain = self:GetPosition(self.MainForm)
 -- Do task for size change in the move event as workaround to prevent consuming size change events.
 local grid = self.MainForm:FindChild("AttributeGrid")
 grid:SetColumnWidth(2, grid:GetWidth() - 35 - 100)
end

--****************************************************************************************************
-- TargetSpy Helper Functions
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- TargetSpy GetAttributes
-- This function collect all attributes that avaiable for targets and prepare data for show in the
-- spy window.
------------------------------------------------------------------------------------------------------
function TargetSpy:GetAttributes(target)
 if target ~= nil then
  local attributes =
  {
   {
-- Template for attribute item.
--    {
--     name = Apollo.GetString(""),
--     value = math.ceil(target:()),
--     icon = ""
--    },
    -- Block for general attributes.
    {
     name = Apollo.GetString("Character_MaxHealthLabel"),
     value = TargetSpy:MathWrapperCeil(target:GetMaxHealth()),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_Health"
    },
    {
     name = Apollo.GetString("Character_MaxShieldLabel"),
     value = TargetSpy:MathWrapperCeil(target:GetShieldCapacityMax()),
     icon = ""
    },
    {
     name = Apollo.GetString("AttributeAssaultPower"),
     value = TargetSpy:MathWrapperCeil(target:GetAssaultPower()),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_AssaultPower"
    },
    {
     name = Apollo.GetString("AttributeSupportPower"),
     value = TargetSpy:MathWrapperCeil(target:GetSupportPower()),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_SupportPower"
    },
   },
   -- Block with mount infos.
   {
    {
     name = "Max Mount Health",
     value = TargetSpy:MathWrapperCeil(target:GetMountMaxHealth()),
     icon = ""
    }
   },
   -- Block with secondary attributes.
   { -- Critical Hit informations.
    {
     name = Apollo.GetString("Character_CritSeverityLabel"),
     value = self:CombatWrapper(target:GetCritSeverity(), 1),
     --value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetCritSeverity() + 0.000005) * 10000) / 100)),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_CriticalSeverity"
    }
   },
   { -- Shield informations.
    {
     name = Apollo.GetString("Character_ShieldRegenPercentLabel"),
     value = self:CombatWrapper(target:GetShieldRegenPct(), 1),
     --value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetShieldRegenPct() + 0.000005) * 10000) / 100)),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_ShieldRebootLabel"),
     value = self:CombatWrapper(target:GetShieldRebootTime(), 2),
     --value = String_GetWeaselString(Apollo.GetString("Character_SecondsLabel"), target:GetShieldRebootTime() / 1000),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_ShieldRegenIntervalLabel"),
     value = self:CombatWrapper(target:GetShieldTickTime(), 3),
     --value = String_GetWeaselString(Apollo.GetString("Character_PerSecLabel"), target:GetShieldTickTime() / 1000),
     icon = ""
    }
   },
   { -- Focus informations.
    {
     name = Apollo.GetString("Character_ManaRecoveryLabel"),
     value = self:CombatWrapper(target:GetManaRegenInCombat(), 4),
     --value = String_GetWeaselString(Apollo.GetString("Character_PerSecLabel"), target:GetManaRegenInCombat() * 2),
     icon = ""
    }
   },
   { -- PvP informations.
    {
     name = Apollo.GetString("Character_PvPOffenseLabel"),
     value = TargetSpy:PvPPercentWrapper(target:GetPvPOffenseRating(), target:GetPvPOffensePercent()),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_PvPOffenseLabel") .. " Rating",
     value = self:CombatWrapper(target:GetPvPOffenseRating()),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_PvPDefLabel"),
     value = TargetSpy:PvPPercentWrapper(target:GetPvPDefenseRating(), target:GetPvPDefensePercent()),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_PvPDefLabel") .. " Rating",
     value = self:CombatWrapper(target:GetPvPDefenseRating()),
     icon = ""
    }
   }
  }
  return attributes
 end
end

------------------------------------------------------------------------------------------------------
-- TargetSpy GetAttributesSelf
-- This function collect all attributes that avaiable for self and prepare data for show in the spy window.
------------------------------------------------------------------------------------------------------
function TargetSpy:GetAttributesSelf(target)
 if target ~= nil then
  local attributes =
  {
   {
    -- Block for general attributes.
    {
     name = Apollo.GetString("Character_MaxHealthLabel"),
     value = TargetSpy:MathWrapperCeil(target:GetMaxHealth()),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_Health"
    },
    {
     name = Apollo.GetString("Character_MaxShieldLabel"),
     value = TargetSpy:MathWrapperCeil(target:GetShieldCapacityMax()),
     icon = ""
    },
    {
     name = Apollo.GetString("AttributeAssaultPower"),
     value = TargetSpy:MathWrapperCeil(target:GetAssaultPower()),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_AssaultPower"
    },
    {
     name = Apollo.GetString("AttributeSupportPower"),
     value = TargetSpy:MathWrapperCeil(target:GetSupportPower()),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_SupportPower"
    },
    -- The function GetUnitProperties() return nil for a target, so that Armor information and Primary Attributes can not collected in this way.
    {
     name = Apollo.GetString("CRB_Armor"),
     value = TargetSpy:MathWrapperCeil(target:GetUnitProperties().Armor.fValue),
     icon = ""
    }
   },
   -- Block with mount infos.
   {
    {
     name = "Max Mount Health",
     value = TargetSpy:MathWrapperCeil(target:GetMountMaxHealth()),
     icon = ""
    }
   },
   -- Block with secondary attributes.
   { -- Critical Hit informations.
    {
     name = Apollo.GetString("Character_StrikethroughLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetStrikethroughChance() + 0.000005) * 10000) / 100)),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_Strikethrough"
    },
    {
     name = Apollo.GetString("Character_CritChanceLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetCritChance() + 0.000005) * 10000) / 100)),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_CriticalHit"
    },
    {
     name = Apollo.GetString("Character_CritSeverityLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetCritSeverity() + 0.000005) * 10000) / 100)),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_CriticalSeverity"
    },
    {
     name = Apollo.GetString("Character_ArmorPenLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetIgnoreArmorBase() + 0.000005) * 10000) / 100)),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_ShieldPenLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetIgnoreShieldBase() + 0.000005) * 10000) / 100)),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_LifestealLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetBaseLifesteal() + 0.000005) * 10000) / 100)),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_HasteLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (-1 * math.floor((target:GetCooldownReductionModifier() + 0.000005 - 1) * 10000) / 100)),
     icon = ""
    }
   },
   { -- Shield informations.
    {
     name = Apollo.GetString("Character_ShieldRegenPercentLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetShieldRegenPct() + 0.000005) * 10000) / 100)),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_ShieldRebootLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_SecondsLabel"), target:GetShieldRebootTime() / 1000),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_ShieldRegenIntervalLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PerSecLabel"), target:GetShieldTickTime() / 1000),
     icon = ""
    }
   },
   { -- Resistance informations.
    {
     name = Apollo.GetString("Character_PhysicalMitLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetPhysicalMitigation() + 0.000005) * 10000) / 100)),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_TechMitLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetTechMitigation() + 0.000005) * 10000) / 100)),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_MagicMitLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetMagicMitigation() + 0.000005) * 10000) / 100)),
     icon = ""
    }
   },
   { -- Deflect informations.
    {
     name = Apollo.GetString("Character_DeflectLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetDeflectChance() + 0.000005) * 10000) / 100)),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_Deflect"
    },
    {
     name = Apollo.GetString("Character_DeflectCritLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((target:GetDeflectCritChance() + 0.000005) * 10000) / 100)),
     icon = "ClientSprites:Icon_Windows_UI_CRB_Attribute_DeflectCritical"
    },
    {
     name = Apollo.GetString("Character_ResilianceLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((math.abs(target:GetCCDurationModifier() -1) + 0.000005) * 10000) / 100)),
     icon = ""
    }
   },
   { -- Focus informations.
    {
     name = Apollo.GetString("Character_ManaRecoveryLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PerSecLabel"), target:GetManaRegenInCombat() * 2),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_ManaCostRedLabel"),
     value = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((math.abs(target:GetManaCostModifier() -1) + 0.000005) * 10000) / 100)),
     icon = ""
    }
   },
   { -- PvP informations.
    {
     name = Apollo.GetString("Character_PvPOffenseLabel"),
     value = TargetSpy:PvPPercentWrapper(target:GetPvPOffenseRating(), target:GetPvPOffensePercent()),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_PvPOffenseLabel") .. " Rating",
     value = target:GetPvPOffenseRating(),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_PvPDefLabel"),
     value = TargetSpy:PvPPercentWrapper(target:GetPvPDefenseRating(), target:GetPvPDefensePercent()),
     icon = ""
    },
    {
     name = Apollo.GetString("Character_PvPDefLabel") .. " Rating",
     value = target:GetPvPDefenseRating(),
     icon = ""
    }
   }
  }
  return attributes
 end
end

------------------------------------------------------------------------------------------------------
-- TargetSpy GetPosition
-- Returns the position of the form as table.
------------------------------------------------------------------------------------------------------
function TargetSpy:GetPosition(form)
 local result = nil
 if form ~= nil then
  local leftAnchor, topAnchor, rightAnchor, bottomAnchor = form:GetAnchorOffsets()
  result =
  {
   Bottom = bottomAnchor,
   Left = leftAnchor,
   Right = rightAnchor,
   Top = topAnchor,
  }
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- TargetSpy MathWrapperCeil
-- Wrapper for math function ceil to handle nil
------------------------------------------------------------------------------------------------------
function TargetSpy:MathWrapperCeil(value)
 local result = 0
 if value ~= nil then
  result = math.ceil(value)
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- TargetSpy MathWrapperCeil
-- Wrapper for some functions, that gives no result if player self is in combat. Avaiable modes are:
-- 1 for Character_PercentAppendLabel - math.floor((value + 0.000005) * 10000) / 100)
-- 2 for Character_SecondsLabel - value / 1000
-- 3 for Character_PerSecLabel - value / 1000
-- 4 for Character_PerSecLabel - value * 2
-- 5 for value
------------------------------------------------------------------------------------------------------
function TargetSpy:CombatWrapper(value, mode)
 local result = "n/a"
 if value ~= nil then
  if mode == 1 then
   result = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((value + 0.000005) * 10000) / 100))
  elseif mode == 2 then
   result = String_GetWeaselString(Apollo.GetString("Character_SecondsLabel"), value / 1000)
  elseif mode == 3 then
   result = String_GetWeaselString(Apollo.GetString("Character_PerSecLabel"), value / 1000)
  elseif mode == 4 then
   result = String_GetWeaselString(Apollo.GetString("Character_PerSecLabel"), value * 2)
  elseif mode == 5 then
   result = value
  end
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- TargetSpy MainHide
-- Close the main form and deregister events.
------------------------------------------------------------------------------------------------------
function TargetSpy:MainHide()
 -- Set config that main view is not visible.
 self.Config.IsVisible = false
 -- Unregister event.
 Apollo.RemoveEventHandler("TargetUnitChanged", self)
 -- Hide windows.
 self.MainForm:Show(false, false)
end

------------------------------------------------------------------------------------------------------
-- TargetSpy MainShow
-- Open the main form and do preparations.
------------------------------------------------------------------------------------------------------
function TargetSpy:MainShow()
 -- Set config that main view is visible.
 self.Config.IsVisible = true
 -- Register event again and update target infos.
 Apollo.RegisterEventHandler("TargetUnitChanged", "OnTargetUnitChanged", self)
 self:OnTargetUnitChanged(GameLib.GetTargetUnit())
 self.MainForm:Show(true, false)
end

------------------------------------------------------------------------------------------------------
-- TargetSpy PvPPercentWrapper
-- Wrapper for PvP Values, to get a apropriate value.
------------------------------------------------------------------------------------------------------
function TargetSpy:PvPPercentWrapper(rating, percent)
 local result = "n/a"
 if rating ~= 0 and percent ~= nil then
  result = String_GetWeaselString(Apollo.GetString("Character_PercentAppendLabel"), (math.floor((percent + 0.000005) * 10000) / 100))
 end
 return result
end

--****************************************************************************************************
-- TargetSpy Instance
--****************************************************************************************************
local TargetSpyInst = TargetSpy:new()
TargetSpyInst:Init()
